/*===========================================================*\
    This class models a thread waits for an external device
    which signals its readiness via Win32 events.  To simulate
    such a device, we are using Win32 waitable timers.
\*===========================================================*/
#include "DeviceThread.h"
#include "Miscellaneous.h"


// Simple "virtual" device class for simulating an external
// physical device. This class simulates devices having 
// semantics that signal via Win32 synchronization handles, 
// such as events.  For purposes of simulation, this class
// substitutes a waitable timer handle for an event handle.
class Device
{
  public:
    Device() : handle_(::CreateWaitableTimer(NULL, true, NULL))
    {}

    ~Device()   {::CloseHandle(handle_);}

    void StartOperation(__int64 expectedDuration_ms)
    {
        LARGE_INTEGER   time;
        time.QuadPart = expectedDuration_ms * -10000;

        BOOL result = ::SetWaitableTimer(
            handle_, &time, 0, NULL, NULL, true
        );

        DWORD err = ::GetLastError();
    }

    void StopOperation()
    {
        BOOL status = ::CancelWaitableTimer(handle_);
    }

    // Caller can treat this handle as an event handle,
    // and wait for it to be signalled like an event.
    HANDLE  Handle() const  {return handle_;}
    
  private:
    const HANDLE    handle_;
};




// Body of thread here
int DeviceThread::Run()
{
    Device  myDevice;

    // If using a mediator, the code must try/catch the
    // exception type thrown by the mediator.
    try
    {
        while (true)
        {
            myDevice.StartOperation(RandomDelay());
            Print(Name() + " started");

            _mediator.Wait( 
                myDevice.Handle(),
                EventSynchronizer_t::TIME_INFINITE 
            );

            myDevice.StopOperation();
            Print(Name() + " completed");
        }
    }
    catch(Mediator_t::Exception_t &)
    {
        // Thread was interrupted by mediator; exit gracefully
        Print(Name() + " interrupted");
    }

    return 0;
}

