/*===========================================================*\
\*===========================================================*/
#if !defined(__SOCKET_SYNCHRONIZER_H__)
#define __SOCKET_SYNCHRONIZER_H__

#include <windows.h>

namespace Tx
{
    template <class Exception, class Mutex>
    class SocketSynchronizer
    {
      public:
        typedef SOCKET      SyncObject_t; // What to wait on
        typedef Exception   Exception_t;
        typedef long        TimeDuration_t;
        typedef Mutex       Mutex_t;

        enum {TIME_INFINITE = -1};

        SocketSynchronizer()                
        {
            Reset();
        }

        virtual ~SocketSynchronizer()       
        {
            Signal();
        }

        //-----------------------------------------------------
        // Blocking interruptible wait for syncObject
        bool Wait(  const SyncObject_t & syncObject, 
                    TimeDuration_t timeout
                ) const 
        {
            SyncObject_t    handles[2];
            handles[0] = _signalObject;
            handles[1] = syncObject;
            return WaitImpl(handles, 2, timeout);
        }

        //-----------------------------------------------------
        // Blocking idle sleep that is interruptible
        void Sleep(TimeDuration_t timeout) const    
        {
            WaitImpl(&_signalObject, 1, timeout);
        }

        void Signal()
        {
            ::closesocket(_signalObject);
        }

        void Reset()
        {
            _signalObject = ::socket(AF_INET, SOCK_DGRAM, 0);
        }

      private:
        //-----------------------------------------------------
        // Actual implementation of "multiple-listening" wait
        // behavior relying on select().
        bool WaitImpl(  
                const SyncObject_t * handles,
                int numHandles,
                TimeDuration_t timeout  // ms or TIME_INFINITE
            ) const
        {
            const int MS_PER_SEC = 1000;
            const int US_PER_MS = 1000;

            fd_set readfds;

            FD_ZERO(&readfds);

            FD_SET(handles[0], &readfds);
            FD_SET(handles[1], &readfds);

            struct timeval  timev;
            struct timeval * ptimev = NULL; 

            if (timeout != TIME_INFINITE)
            {
                timev.tv_sec = timeout / MS_PER_SEC;
                timev.tv_usec = 
                    (timeout % MS_PER_SEC) * US_PER_MS;
                ptimev = &timev;
            }

            // Wait on both sockets simultaneously
            int status = ::select(NULL, &readfds, NULL, NULL, ptimev);

            if (SelectFailed(status))
            {
                // Should include more robust err handling here
                return false;
            }
            else if (status == 0)
            {
                // timed out
                return false;   // Nothing happened
            }
            else if (FD_ISSET(handles[0], &readfds))
            {
                throw Exception_t();    // Thread interrupted
            }
            else if (FD_ISSET(handles[1], &readfds))
            {
                return true;    // Desired event is signaled
            }
            else
            {
                // Should include more robust err handling here
                return false;
            }
        }

        // Wrapper to abstract the differences between
        // Win32 select() and the rest of the world
        static inline bool SelectFailed(int status)
        {
            #ifdef WIN32
                return (status == SOCKET_ERROR);
            #else
                return (status < 0);
            #endif
        }

        // Disallow copy constructor/assignment operators
        SocketSynchronizer(const SocketSynchronizer &);
        SocketSynchronizer & operator=(
                const SocketSynchronizer &);

        SyncObject_t        _signalObject;
    };
}

#endif // __SOCKET_SYNCHRONIZER_H__
