/*
 * path.c
 *
 * $Id: path.c,v 1.4 2003/02/12 10:52:57 danny Exp $
 *
 * THE BEER-WARE LICENSE
 *
 * <danny@alphazed.com> wrote this software. As long as you retain this notice,
 * you can do whatever you want with this stuff.  If we meet some day, and you
 * think this stuff is worth it, you can buy me a beer in return.
 *
 * Daniel Lawrence
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <toolbox.h>
#include "readdb.h"
#include "conf.h"
#include "res.h"
#include "sym.h"

typedef enum {
	SHELL_SH,
	SHELL_CSH
} shell_t;


static void exit_mem(void);
static void exit_nohome(void);
static void exit_noshell(void);
static void exit_msg(char *);
static int exist(char *);
static void output(int, int, RESULT *);


enum {
	MSG_NONE,
	MSG_USAGE,
	MSG_ERROR
};


int
main(int argc, char **argv)
{
	char *home;
	char *shell;
	char *host;
	char *conf = NULL;		/* $HOME/.path */
	char *database = NULL;		/* .../etc/path.db */
	shell_t shelltype;
	int c;
	FILE *f;
	char *msg = NULL;
	RESULT r;
	int help = MSG_NONE;
	int use_tmp = 0;

	/* atexit(tb_strdump); */

	tb_setprogname(argv[0]);

	/*
	 * $HOME and $SHELL must exist and be non-empty
	 */
	if ((home = getenv("HOME")) == NULL || home[0] == '\0')
		exit_nohome();
	if ((shell = getenv("SHELL")) == NULL || home[0] == '\0')
		exit_noshell();

	/*
	 * default shelltype based on $SHELL
	 */
	shelltype = (strstr(shell, "csh") == NULL) ? SHELL_SH : SHELL_CSH;

	if ((host = tb_gethostname()) == NULL)
		exit_mem();

	/*
	 * parse command line
	 */
	while ((c = getopt(argc, argv, "c:d:hs:t")) != -1)
		switch (c) {
		case 'c':
			if ((conf = tb_strcpy(conf, optarg)) == NULL)
				exit_mem();
			break;
		case 'd':
			if ((database = tb_strcpy(database, optarg)) == NULL)
				exit_mem();
			break;
		case 'h':
			help = MSG_USAGE;
			break;
		case 's':
			if (strcmp(optarg, "csh") == 0)
				shelltype = SHELL_CSH;
			else
				shelltype = SHELL_SH;
			break;
		case 't':
			use_tmp = 1;
			break;
		default:
			help = MSG_ERROR;
			break;
		}

	/*
	 * set any remaining defaults
	 */
	/* if ~/.path.<hostname> exists, use it. otherwise use ~/.path */
	if (conf == NULL) {
		conf = tb_strprintf(conf, "%s/.path.%s", home, host);
		if (!exist(conf))
	        	conf = tb_strprintf(conf, "%s/.path", home);
	}
	if (database == NULL)
	        database = tb_strprintf(database, "%s/etc/path.db", BASE);
	if (conf == NULL || database == NULL)
		exit_mem();

	/*
	 * display usage and default paths if -h or command line error
	 */
	if (help != MSG_NONE) {
		fprintf(
			stderr,
			"usage: %s [-h] [-t] [-s sh|csh] [-c conffile] [-d database]\n",
			tb_getprogname()
		);

		printf("conffile=%s\n", conf);
		printf("database=%s\n", database);
		printf(
			"shelltype=%s\n",
			(shelltype == SHELL_CSH) ? "csh" : "sh"
		);
                printf("http://www.alphazed.co.uk/software/path/\n");
		exit(help == MSG_ERROR);
	}


	/*
	 * load database
	 */
	if ((f = fopen(database, "r")) == NULL) {
		msg = tb_strprintf(
			msg,
			"can't open database \"%s\": %s",
			database,
			strerror(errno)
		);
		if (msg == NULL)
			exit_mem();
		else
			exit_msg(msg);
	}
	if ((msg = readdb(f, msg)) == NULL)
		exit_mem();
	else if (strlen(msg))
		exit_msg(msg);
	fclose(f);

	if (!res_init(&r))
		exit_mem();

	/*
	 * open conf file and build result
	 */
	if ((f = fopen(conf, "r")) == NULL)
		msg = sym_eval("default", &r, msg);
	else {
		msg = conf_eval(f, &r, msg);
		fclose(f);
	}
	if (msg == NULL)
		exit_mem();
	else if (strlen(msg))
		exit_msg(msg);

	/*
	 * output result
	 */
	output(use_tmp, shelltype, &r);

	exit(0);
}


static void
exit_mem(void)
{
	fprintf(stderr, "%s: out of memory\n", tb_getprogname());
	exit(1);
}


static void
exit_nohome(void)
{
	fprintf(stderr, "%s: HOME must be set\n", tb_getprogname());
	exit(1);
}


static void
exit_noshell(void)
{
	fprintf(stderr, "%s: SHELL must be set\n", tb_getprogname());
	exit(1);
}


static void
exit_msg(char *msg)
{
	fprintf(stderr, "%s: %s\n", tb_getprogname(), msg);
	exit(1);
}


static int
exist(char *path)
{
	/* true if path exists */

	struct stat sb;

	return !(stat(path, &sb) == -1);
}


static void
output(int use_tmp, int shelltype, RESULT *rp)
{
	static char *tmpdir = NULL;	/* not dynamic */
	static char *prefix = NULL;
	static char *path = NULL;
	FILE *f;
	int fd;
	int err;

        if (!use_tmp)
                f = stdout;
        else {
		if (tmpdir == NULL) {
			tmpdir = getenv("TMPDIR");
			if (tmpdir == NULL || tmpdir[0] == '\0')
				tmpdir = "/tmp";
		}
		prefix = tb_strprintf(
			prefix,
			"%s/path.",
			tmpdir
		);
		if (prefix == NULL)
			exit_mem();
		if ((fd = tb_mktmp(prefix, &path, &err)) == -1) {
			if (err == TB_EMEM)
				exit_mem();
			/* should print better error message */
			exit_msg("can't create temp file");
		}
		if ((f = fdopen(fd, "w")) == NULL) {
			/* should print better error message */
			exit_msg("fdopen");
		}
        }
        if (shelltype == SHELL_SH)
                res_print_sh(f, rp);
        else
                res_print_csh(f, rp);
        if (use_tmp) {
                fclose(f);
                printf("%s\n", path);
        }
}
