/*
 * sym.c -- symbol table
 *
 * $Id: sym.c,v 1.2 2003/02/11 16:12:34 danny Exp $
 *
 * THE BEER-WARE LICENSE
 *
 * <danny@alphazed.com> wrote this software. As long as you retain this notice,
 * you can do whatever you want with this stuff.  If we meet some day, and you
 * think this stuff is worth it, you can buy me a beer in return.
 *
 * Daniel Lawrence
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <toolbox.h>
#include "list.h"
#include "var.h"
#include "gettok.h"
#include "res.h"
#include "sym.h"


typedef enum {
	ENTRY_T,
	MACRO_T
} entry_t;

typedef struct {
	char *buf;	/* dynamic string holding all data */
	char *name;
	entry_t t;	/* entry or macro */

	/* ENTRY_T */
	char *path;
	char *manpath;
	VAR v;

	/* MACRO_T */
	LIST l;
} SYMBOL;


/* symbol table; simple array */
static SYMBOL *tab = NULL;
static int tabsize = 0;
static int tabused = 0;


static SYMBOL *
sym_lookup(char *name)
{
	int i;

	for (i = 0; i < tabused; ++i)
		if (strcmp(tab[i].name, name) == 0)
			return tab + i;
	return NULL;
}


/*
 * parse_macro -- parse macro expansion from line
 */
static char *
parse_macro(char *line, SYMBOL *p, char *msg)
{
	char *tok;

	while ((tok = gettok_quoted(&line)) != NULL)
		if (!list_add(&p->l, tok))
			return tb_strfree(msg);
	return tb_strcpy(msg, "");
}


/*
 * parse_entry -- parse non-macro entry from line
 */
static char *
parse_entry(char *line, SYMBOL *p, char *msg)
{
	char *var;
	char *val;

	/* for later convenience we want path and manpath
	 * to at least be an empty string instead of NULL
	 */
	if ((p->path = gettok_quoted(&line)) == NULL) {
		p->path = line;		/* \0 */
		p->manpath = line;	/* \0 */
		return tb_strcpy(msg, "");
	}
	if (strcmp(p->path, "-") == 0)
		p->path[0] = '\0';

	if ((p->manpath = gettok_quoted(&line)) == NULL) {
		p->manpath = line;	/* \0 */
		return tb_strcpy(msg, "");
	}
	if (strcmp(p->manpath, "-") == 0)
		p->manpath[0] = '\0';

	for (;;) {
		if ((var = gettok_quoted(&line)) == NULL)
			return tb_strcpy(msg, "");
		if ((val = gettok_quoted(&line)) == NULL)
			return tb_strcpy(msg, "missing env value");
		if (!var_add(&p->v, var, val))
			return tb_strfree(msg);
	}
	/* can't happen */
	return tb_strfree(msg);
}


/* squeeze first char (@) out of string */
static char *
squeeze(char *s)
{
	char *cp;

	for (cp = s; *(cp+1); ++cp)
		*cp = *(cp+1);
	*cp = '\0';
	return s;
}


/*
 * parse_line -- parse logical line from file; init symbol struct
 */
static char *
parse_line(char *line, SYMBOL *p, char *msg)
{
	char *cp;

	/* make working copy of logical line */
	if ((p->buf = tb_strcpy(NULL, line)) == NULL)
		return tb_strfree(msg);

	/* get first token on logical line */
	cp = p->buf;
	if ((p->name = gettok_white(&cp)) == NULL) {
		tb_strfree(p->buf);
		return tb_strcpy(msg, "missing name");
	}

	if (p->name[0] == '@') {
		/* squeeze out leading '@' */
		if (squeeze(p->name)[0] == '\0') {
			tb_strfree(p->buf);
			return tb_strcpy(msg, "missing macro name");
		}
		p->t = MACRO_T;
		list_init(&p->l);
		msg = parse_macro(cp, p, msg);
	} else {
		p->t = ENTRY_T;
		var_init(&p->v);
		p->path = NULL;
		p->manpath = NULL;
		msg = parse_entry(cp, p, msg);
	}
	if (msg == NULL || msg[0] != '\0') {
		/* cleanup p */
		if (p->t == MACRO_T)
			list_free(&p->l);
		else
			var_free(&p->v);
		tb_strfree(p->buf);
	}
	return msg;
}


char *
sym_parse(char *line, char *msg)
{
	SYMBOL *newp;

	if (tabsize == 0) {
		if ((tab = (SYMBOL *)malloc(sizeof(SYMBOL))) == NULL)
			return tb_strfree(msg);
		tabsize = 1;
	} else if (tabused >= tabsize) {
		if ((newp = (SYMBOL *)realloc(tab, 2 * sizeof(SYMBOL) * tabsize)) == NULL)
			return tb_strfree(msg);
		tabsize *= 2;
		tab = newp;
	}
		
	msg = parse_line(line, &tab[tabused], msg);
	if (msg && msg[0] == '\0')
		++tabused;
	return msg;
}


void
sym_print(void)
{
	int i;
	int j;
	int n;
	char *var;
	char *val;

	for (i = 0; i < tabused; ++i) {
		if (tab[i].t == MACRO_T) {
			printf("%s: MACRO\n", tab[i].name);
			n = list_size(&tab[i].l);
			for (j = 0; j < n; ++j)
				printf("\t%s\n", list_get(&tab[i].l, j));
		} else {
			printf("%s: ENTRY\n", tab[i].name);
			printf("\tpath: %s\n", tab[i].path);
			printf("\tmanpath: %s\n", tab[i].manpath);
			n = var_size(&tab[i].v);
			for (j = 0; j < n; ++j) {
				var_get(&tab[i].v, j, &var, &val);
				printf("\t%s=%s\n", var, val);
			}
		}
	}
}


char *
sym_eval(char *name, RESULT *res, char *msg)
{
	SYMBOL *s;
	int n;
	int i;
	char *var;
	char *val;
	char *err;
	char *word;

	if ((s = sym_lookup(name)) == NULL) {
		/* symbol not found; ok, just ignore */
	} else if (s->t == MACRO_T) {
		n = list_size(&s->l);
		for (i = 0; i < n; ++i) {
			word = list_get(&s->l, i);
			err = sym_eval(word, res, msg);
			if (err == NULL || err[0] != '\0')
				return err;
		}
	} else {
		/* add to list of env vars */
		n = var_size(&s->v);
		for (i = 0; i < n; ++i) {
			var_get(&s->v, i, &var, &val);
			if (!res_env(res, var, val))
				return tb_strfree(msg);
		}
		if (!res_path(res, s->path))
			return tb_strfree(msg);
		if (!res_manpath(res, s->manpath))
			return tb_strfree(msg);
	}
	return tb_strcpy(msg, "");
}
