/*
 * miniinstall.c -- portable installer
 *
 * $Id: miniinstall.c,v 1.4 2003/02/11 16:01:45 danny Exp $
 *
 * THE BEER-WARE LICENSE
 *
 * <danny@alphazed.com> wrote this software. As long as you retain this notice,
 * you can do whatever you want with this stuff.  If we meet some day, and you
 * think this stuff is worth it, you can buy me a beer in return.
 *
 * Daniel Lawrence
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include "toolbox.h"

static void usage(void);
static void mkdirs(int, char **);
static void files(int, char **);
static void file2file(char *, char *);
static void file2dir(char *, char *);
static int perms(char *, uid_t, gid_t, mode_t, char **);

static uid_t uid = (uid_t)-1;
static gid_t gid = (gid_t)-1;
static mode_t mode = (mode_t)-1;

int
main(int argc, char **argv)
{
	int do_mkdir = 0;
	int c;

	tb_setprogname(argv[0]);

	while ((c = getopt(argc, argv, "dg:m:o:")) != -1)
		switch (c) {
		case 'd':
			do_mkdir = 1;
			break;
		case 'g':
			if (!tb_gidof(optarg, &gid))
				tb_die_msg("invalid group");
			break;
		case 'm':
			if (!tb_modeof(optarg, &mode))
				tb_die_msg("invalid mode");
			break;
		case 'o':
			if (!tb_uidof(optarg, &uid))
				tb_die_msg("invalid user");
			break;
		default:
			usage();
		}
	argc -= optind;
	argv += optind;

	if (argc < 1)
		usage();

	if (do_mkdir)
		mkdirs(argc, argv);
	else
		files(argc, argv);

	exit(0);
}


static void
usage(void)
{
	fprintf(
		stderr,
		"usage: %s [-d] [-o owner] [-g group] [-m mode] file [...] dest\n",
		tb_getprogname()
	);
	exit(1);
}



static void
mkdirs(int argc, char **argv)
{
	static char *msg = NULL;

	while (argc-- > 0)
		if (!tb_mkdirp(*argv++, uid, gid, mode, &msg))
			tb_die_msg(msg);
}


static void
files(int argc, char **argv)
{
	struct stat sb;

	if (argc < 2)
		usage();
	if (stat(argv[argc-1], &sb) == -1) {
		if (errno != ENOENT)
			tb_die_printf("can't stat %s:", argv[argc-1]);
		else if (argc != 2)
			usage();
		else
			file2file(argv[0], argv[1]);
	} else if (sb.st_mode & S_IFDIR) {
		while (argc > 1) {
			file2dir(*argv, argv[argc-1]);
			--argc;
			++argv;
		}
	} else if (!(sb.st_mode & S_IFREG)) {
		tb_die_printf(
			"%s: destination is not a directory or a file",
			argv[argc-1]
		);
	} else if (argc != 2)
		usage();
	else
		file2file(argv[0], argv[1]);
}


static void
file2file(char *src, char *dst)
{
	static char *tmppath = NULL;
	static char *msg = NULL;
	struct stat sb;
	int s;
	int d;
	int err;

	/* verify src is a regular file */
	if (stat(src, &sb) == -1)
		tb_die_printf("%s: can't stat:", src);
	if (!(sb.st_mode & S_IFREG))
		tb_die_printf("%s not a regular file", src);

	if ((s = open(src, O_RDONLY)) == -1)
		tb_die_printf("%s: can't open:", src);

	if ((d = tb_mktmp(dst, &tmppath, &err)) == -1) {
		if (err == TB_EMEM)
			tb_die_msg("out of memory");
		else
			tb_die_printf("can't create temporary file:");
	}

	if (!tb_cp(s, d, &msg)) {
		unlink(tmppath);
		tb_die_msg(msg);
	}

	if (close(s) == -1) {
		unlink(tmppath);
		tb_die_printf("%s: close error:", src);
	}
	if (close(d) == -1) {
		unlink(tmppath);
		tb_die_printf("%s: close error:", tmppath);
	}

	if (!perms(tmppath, uid, gid, mode, &msg)) {
		unlink(tmppath);
		tb_die_msg(msg);
	}

	if (rename(tmppath, dst) == -1) {
		unlink(tmppath);
		tb_die_printf("can't rename %s -> %s:", tmppath, dst);
	}
}


static void
file2dir(char *src, char *dst)
{
	static char *dstpath = NULL;
	char *name;

	if ((name = strrchr(src, '/')) == NULL)
		name = src;
	else
		++name;

	dstpath = tb_strprintf(
		dstpath,
		"%s/%s",
		dst,
		name
	);
	if (dstpath == NULL)
		tb_die_msg("out of memory");

	file2file(src, dstpath);
}


static int
perms(char *path, uid_t uid, gid_t gid, mode_t mode, char **msgp)
{
	struct stat sb;

	if (stat(path, &sb) == -1) {
		*msgp = tb_strprintf(
			*msgp,
			"%s: can't stat:",
			path
		);
		return 0;
	}

	if ((uid != (uid_t)-1 || gid != (gid_t)-1)
	    && (uid != sb.st_uid || gid != sb.st_gid)) {
		if (chown(path, uid, gid) == -1) {
			*msgp = tb_strprintf(
				*msgp,
				"%s: can't chown:",
				path
			);
			return 0;
		}
	}

	if (mode != (mode_t)-1 && mode != sb.st_mode) {
		if (chmod(path, mode) == -1) {
			*msgp = tb_strprintf(
				*msgp,
				"%s: can't chmod:",
				path
			);
			return 0;
		}
	}

	return 1;
}
