/*
 * salt.c -- generate random salt
 *
 * state machine version
 *
 * $Id: salt.c,v 1.3 2003/03/10 11:08:42 danny Exp $
 *
 * THE BEER-WARE LICENSE
 *
 * <danny@alphazed.com> wrote this software. As long as you retain this notice,
 * you can do whatever you want with this stuff.  If we meet some day, and you
 * think this stuff is worth it, you can buy me a beer in return.
 *
 * Daniel Lawrence
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "toolbox.h"

/* both lexer and parser states are listed here */
enum {
	STATE_START,		/* idle state */
	STATE_IN_GENERATOR,	/* between [ and ] */
	STATE_SAW_FIRST,	/* saw first char of possible range */
	STATE_SAW_THROUGH,	/* saw the - in a range */
	STATE_SAW_GENERATOR,	/* saw the ] of a generator */
	STATE_IN_COUNT,		/* saw {, accumulating count */
	STATE_SAW_COUNT,	/* saw } */
	STATE_SAW_ESCAPE,	/* for gettok() */
	STATE_IN_OCTAL,		/* for gettok() */
	STATE_EOF
};

/* token types returned by lexer */
enum {
	TOK_ERR,		/* error in lexer */
	TOK_EOF,
	TOK_DIGIT,		/* 0-9 */
	TOK_GEN_START,		/* [ */
	TOK_GEN_END,		/* ] */
	TOK_THROUGH,		/* - */
	TOK_COUNT_START,	/* { */
	TOK_COUNT_END,		/* } */
	TOK_LITERAL,		/* a normal character */
	TOK_EMPTY		/* empty pushback indicator */
};


/* output of lexer is a token */
struct token {
	int type;
	int val;
};

/* parser pushback buffer */
static struct token pushback = { TOK_EMPTY, 0 };

/* current input buffer; never written to */
static char *input;

/* points to next char to be read from input */
static char *ptr;

/* current random character selector; returns a random number [0..n-1] */
static int (*pick)(int);

/* error message buffer */
static char *msg = NULL;


/* default random character selector */
static int
pick_default(int n)
{
	return rand() % n;
}


/* true if c is an octal digit */
static int
isoct(int c)
{
	switch (c) {
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
		return 1;
	default:
		return 0;
	}
}


/* get next char from input */
static int
getnext(void)
{
	return (*ptr == '\0') ? '\0' : *ptr++;
}


/* move input pointer back one character */
static void
unget(void)
{
	if (ptr == input) {
		fprintf(stderr, "tb_salt: too many chars pushed back\n");
		abort();
	}
	--ptr;
}


/* read next token from input, returns token type */
static int
gettok(struct token *t)
{
	int state;
	int c;
	int digits_seen = 0;	/* count octal digits in \nnn */
	

	if (pushback.type != TOK_EMPTY) {
		t->type = pushback.type;
		t->val = pushback.val;
		pushback.type = TOK_EMPTY;
		return t->type;
	}

	state = STATE_START;
	for (;;) {
		c = getnext();

		switch (state) {
		case STATE_START:
			if (c == '\0')
				return t->type = TOK_EOF;
			if (c == '\\')
				state = STATE_SAW_ESCAPE;
			else if (c == '[') {
				t->val = '[';
				return t->type = TOK_GEN_START;
			} else if (c == '-') {
				t->val = '-';
				return t->type = TOK_THROUGH;
			} else if (c == ']') {
				t->val = ']';
				return t->type = TOK_GEN_END;
			} else if (c == '{') {
				t->val = '{';
				return t->type = TOK_COUNT_START;
			} else if (c == '}') {
				t->val = '}';
				return t->type = TOK_COUNT_END;
			} else {
				t->val = c;
				return t->type = TOK_LITERAL;
			}
			break;

		case STATE_SAW_ESCAPE:
			if (c == '\0') {
				t->val = '\\';
				return t->type = TOK_LITERAL;
			}
			if (c == '0' || c == '1' || c == '2' || c == '3') {
				t->val = c - '0';
				digits_seen = 1;
				state = STATE_IN_OCTAL;
			} else {
				/* might as well make it ok to escape
				 * anything
				 */
				t->val = c;
				return t->type = TOK_LITERAL;
			}
			break;

		case STATE_IN_OCTAL:
			if (c == '\0') {
				/* t->val has already been built-up */
				return t->type = TOK_LITERAL;
			}
			if (!isoct(c)) {
				unget();
				return t->type = TOK_LITERAL;
			}
			if (digits_seen == 3) {
				unget();
				/* t->val has already been built up */
				return t->type = TOK_LITERAL;
			}
			++digits_seen;
			t->val = t->val * 8 + c - '0';
			break;

		default:
			msg = tb_strcpy(
				msg,
				"tb_salt: gettok: can't happen"
			);
			return TOK_ERR;
			break;
		}
	}
}


/* push token back on input stream */
void
ungettok(struct token *t)
{
	if (pushback.type != TOK_EMPTY) {
		fprintf(stderr, "tb_salt: too many tokens pushed back\n");
		abort();
	}
	pushback.type = t->type;
	pushback.val = t->val;
}



char *
tb_salt(char *salt, char *template, int (*pick_f)(int))
{
	static char *range = NULL;
	struct token tok;
	int state;
	int first = 0;
	int last = 0;
	int count = 0;
	int i;

        input = (template == NULL) ? "[A-Za-z0-9./]{2}" : template;
        ptr = input;
        pushback.type = TOK_EMPTY;
        pushback.val = 0;
        pick = (pick_f == NULL) ? pick_default : pick_f;
	state = STATE_START;
	if ((salt = tb_strcpy(salt, "")) == NULL)
		return NULL;
	if ((msg = tb_strcpy(msg, "")) == NULL)
		return NULL;

	for (;;) {
		if (gettok(&tok) == TOK_ERR)
			return NULL;

		switch (state) {
		case STATE_START:
			if (tok.type == TOK_EOF)
				return salt;
			/* cheat and allow {, } and ] outside of generators */
			if (tok.type == TOK_LITERAL ||
			    tok.type == TOK_COUNT_START ||
			    tok.type == TOK_COUNT_END ||
			    tok.type == TOK_GEN_END) {
				salt = tb_strcatc(salt, tok.val);
				if (salt == NULL) {
					tb_strfree(msg);
					return NULL;
				}
			} else if (tok.type == TOK_GEN_START) {
				if ((range = tb_strcpy(range, "")) == NULL) {
					msg = tb_strfree(msg);
					return NULL;
				}
				state = STATE_IN_GENERATOR;
			} else {
				msg = tb_strcpy(
					msg,
					"expected literal or '['"
				);
				return NULL;
			}
			break;

		case STATE_IN_GENERATOR:
			if (tok.type == TOK_GEN_END)
				state = STATE_SAW_GENERATOR;
			else if (tok.type == TOK_LITERAL) {
				first = tok.val;
				state = STATE_SAW_FIRST;
			} else {
				msg = tb_strcpy(
					msg,
					"expected literal or ']'"
				);
				return NULL;
			}
			break;

		case STATE_SAW_FIRST:
			if (tok.type == TOK_THROUGH)
				state = STATE_SAW_THROUGH;
			else {
				ungettok(&tok);
				range = tb_strcatc(range, first);
				if (range == NULL) {
					msg = tb_strfree(msg);
					return NULL;
				}
				state = STATE_IN_GENERATOR;
			}
			break;

		case STATE_SAW_THROUGH:
			if (tok.type != TOK_LITERAL) {
				msg = tb_strcpy(
					msg,
					"expected end of range"
				);
				return NULL;
			}
			last = tok.val;
			if (last < first) {
				msg = tb_strcpy(
					msg,
					"backwards range"
				);
				return NULL;
			}
			for (i = first; i <= last; ++i)
				if ((range = tb_strcatc(range, i)) == NULL) {
					msg = tb_strfree(msg);
					return NULL;
				}
			state = STATE_IN_GENERATOR;
			break;

		case STATE_SAW_GENERATOR:
			if (strlen(range) == 0) {
				msg = tb_strcpy(
					msg,
					"empty range"
				);
				return NULL;
			}
			if (tok.type == TOK_COUNT_START) {
				count = 0;
				state = STATE_IN_COUNT;
			} else {
				ungettok(&tok);
				count = 1;
				state = STATE_SAW_COUNT;
			}
			break;

		case STATE_IN_COUNT:
			if (tok.type == TOK_COUNT_END)
				state = STATE_SAW_COUNT;
			else if (tok.type == TOK_LITERAL && isdigit(tok.val))
				count = count * 10 + tok.val - '0';
			else {
				msg = tb_strcpy(
					msg,
					"expected digit or '}'"
				);
				return NULL;
			}
			break;

		case STATE_SAW_COUNT:
			ungettok(&tok);
			for (i = 0; i < count; ++i) {
				salt = tb_strcatc(
					salt,
					range[(pick)(strlen(range))]
				);
				if (salt == NULL) {
					msg = tb_strfree(msg);
					return NULL;
				}
			}
			state = STATE_START;
			break;

		default:
			msg = tb_strcpy(
				msg,
				"tb_salt: can't happen"
			);
			return NULL;
			break;
		}
	}
}


/* return error message */
char *
tb_salt_err(void)
{
	return (msg == NULL) ? "out of memory" : msg;
}
