/**************************************************************************
 * Copyright (c) 2002 Edward Elliott. All rights reserved.
 *
 * Allocation policy to disable swapping.  Uses dlmalloc for allocations
 * and system-specific calls for locking.
 *************************************************************************/

#ifndef NOSWAP_DLMALLOC_H
#define NOSWAP_DLMALLOC_H

#include "misc.h"
#include "dlmalloc.h"

#include <exception>
#include <stdexcept>
#include <errno.h>
#include <string>

// Wrappers around system lock calls
#if defined(_WIN32)
	#include <Windows.h>

	inline int mem_lock(void *addr, size_t len)
	{
		return VirtualLock(addr, len);
	}

	inline int mem_unlock(void *addr, size_t len)
	{
		return VirtualUnlock(addr, len);
	}
#else	// ! _WIN32
	#include <sys/mman.h>

	inline int mem_lock(void *addr, size_t len)
	{
		return mlock(addr, len);
	}

	inline int mem_unlock(void *addr, size_t len)
	{
		return munlock(addr, len);
	}
#endif

/**
 * Allocate with dlmalloc, lock with mem_lock
 */
class NoSwapDLMalloc : public PageTracker, public DummyMaxCount
{
public:
	NoSwapDLMalloc()
	{
		pageSize_ = dlgetpagesize();
	}

	void* doAlloc(size_t numBytes, const void* hint)
	{
		// Try to fulfill the request
		void *pmem = dlmalloc(numBytes);
		if (!pmem)
		{
			throw std::bad_alloc();
		}

		// Lock the result into memory (no swapping)
		if (! mem_lock(pmem, numBytes))
		{
			dlfree(pmem);
			throw std::bad_alloc();
		}

		// Add this block to the page tracker
		addBlock(pmem, numBytes, pageSize_);
		return pmem;
	}

	void doFree(void* pmem, size_t numBytes)
	{
		// Unlock the memory BEFORE it gets freed.
		removeBlock(pmem, numBytes, pageSize_, mem_unlock);
		dlfree(pmem);
	}

private:
	size_t pageSize_;
};

#endif	// NOSWAP_DLMALLOC_H
