/**************************************************************************
 * Copyright (c) 2002 Edward Elliott. All rights reserved.
 *
 * AllocPolicy class using a Windows Heap for allocations and VirtualLock
 * to disable swapping.
 *************************************************************************/

#ifndef NOSWAP_HEAP_H
#define NOSWAP_HEAP_H

#include "misc.h"

#include <Windows.h>
#include <exception>
#include <stdexcept>
#include <string>

/**
 * Allocates memory from a Windows Heap and locks it into memory.  Makes
 * no attempt at thread synchronization, that is done at a higher level.
 */
class NoSwapHeap : public PageTracker, public DummyMaxCount
{
public:
	// Redmond compilers don't allow any member initialization in a class
	// declaration, even static const variables.  Rather than force other
	// compilers to accept this violation of the laws of nature, we'll
	// initialize in-place for all other compilers.  Those using MSVC 
	// either have to include NoSwapHeap.cpp in their projects, or compile
	// it into a lib and link to that (all for 3 measly unsigned longs).
#if defined(_MSC_VER)
	static const DWORD	flags_;		// heap flags
	static const DWORD	initSize_;	// commit a single page
	static const DWORD	maxSize_;	// limit on heap size
#else
	static const DWORD	flags_ = HEAP_NO_SERIALIZE;		// heap flags
	static const DWORD	initSize_ = 1;	// pages to commit on creation
	static const DWORD	maxSize_ = 0;	// limit on heap size
#endif

	NoSwapHeap() : heap_(0), pageSize_(0)
	{
		// Get the current page size
		SYSTEM_INFO sysInfo;
		GetSystemInfo(&sysInfo);
		pageSize_ = sysInfo.dwPageSize;

		// Create a new heap object
		heap_ = HeapCreate(flags_, initSize_, maxSize_);
		if (! heap_)
		{
			throw std::runtime_error(getErrorMsg().c_str());
		}
	}

	void* doAlloc(size_t numBytes, const void* hint)
	{
		// Try to fulfill the request
		void *pmem = HeapAlloc(heap_, flags_, numBytes);
		if (! pmem)
		{
			// HeapAlloc doesn't call SetLastError, no info available
			throw std::bad_alloc();
		}

		// Lock the result into memory (no swapping)
		if (! VirtualLock(pmem, numBytes))
		{
			// Release allocated memory; ignore return value, we already 
			// have an error condition to process
			HeapFree(heap_, flags_, pmem);
			throw std::bad_alloc();
		}

		// Add this block to the page tracker
		addBlock(pmem, numBytes, pageSize_);
		return pmem;
	}

	void doFree(void* pmem, size_t numBytes)
	{
		// Unlock the memory BEFORE it gets freed
		removeBlock(pmem, numBytes, pageSize_, VirtualUnlock);
		if (! HeapFree(heap_, flags_, pmem))
		{
			throw std::runtime_error(getErrorMsg().c_str());
		}
	}

private:
	// Retrieve and format last error code
	std::string getErrorMsg()
	{
		LPTSTR lpMsgBuf;
		FormatMessage( 
			FORMAT_MESSAGE_ALLOCATE_BUFFER | 
			FORMAT_MESSAGE_FROM_SYSTEM | 
			FORMAT_MESSAGE_IGNORE_INSERTS,
			/* msg source */ NULL,
			/* msg id */ GetLastError(),
			/* LanguageId */ 0,
			/* result buf */ (LPTSTR) &lpMsgBuf,
			/* buf size */ 0,
			/* arguments */ NULL 
		);

		std::string msg(lpMsgBuf);
		LocalFree(lpMsgBuf);
		return msg;
	}

	DWORD	pageSize_;	// system page size
	HANDLE	heap_;		// handle to the heap object
};

#endif	// NOSWAP_HEAP_H
