/**************************************************************************
 * Copyright (c) 2002 Edward Elliott. All rights reserved.
 * Scrubbing Policies
 * 
 * Classes for scrubbing (erasing) memory before release.
 *************************************************************************/

#ifndef SCRUBPOL_H
#define SCRUBPOL_H

#include <cstdlib>

/**
 * Helper template to call a generic ScrubPolicy at object destruction
 */
template <class Scrubber>
class ScrubOnDestroy : Scrubber
{
public:
	inline void onDestroy(void* pmem, size_t numBytes)
	{
		doScrub(pmem, numBytes);
	}

	inline void onDeallocate(void* pmem, size_t numBytes)
	{}
};

/**
 * Helper template to call a generic ScrubPolicy on memory deallocation
 */
template <class Scrubber>
class ScrubOnDeallocate : Scrubber
{
public:
	inline void onDestroy(void* pmem, size_t numBytes)
	{}

	inline void onDeallocate(void* pmem, size_t numBytes)
	{
		doScrub(pmem, numBytes);
	}
};

/**
 * Overwrites memory once with a string of zero bytes (0x00).
 * Acceptable the majority of the time.
 */
class BasicScrubber
{
public:
	inline void doScrub(void* pmem, size_t numBytes)
	{
		memset(pmem, 0x00, numBytes);
	}
};

/**
 * Overwrites memory several times with fixed values.
 * Probably not very useful with DRAM.
 */
class HeavyScrubber
{
public:
	inline void doScrub(void* pmem, size_t numBytes)
	{
		memset(pmem, 0xAA, numBytes);
		memset(pmem, 0x55, numBytes);
		memset(pmem, 0xFF, numBytes);
		memset(pmem, 0x00, numBytes);
	}
};

/**
 * Doesn't perform any scrubbing.
 */
class NoScrubber
{
public:
	inline void doScrub(void* pmem, size_t numBytes)	{}
};

#endif	// SCRUBPOL_H
