/**************************************************************************
 * Copyright (c) 2002 Edward Elliott. All rights reserved.
 *
 * Defines the secure allocator template which 
 *************************************************************************/

#ifndef SECALLOC_H
#define SECALLOC_H

#include "misc.h"
#include "AllocPol.h"
#include "ScrubPol.h"
#include "ThreadPol.h"

#include <exception>
#include <stdexcept>

/**
 * Secure Allocator template
 * Combines policies into a concrete STL allocator.
 */
template
<
	class T,
	class ScrubPolicy = ScrubOnDestroy<BasicScrubber>,
	class AllocPolicy = NoSwapAlloc,
	class LockPolicy = NoLocking<AllocPolicy>
>
class SecureAllocator
{
public:
	// Every allocator needs these type definitions
	typedef T			value_type;
	typedef T*			pointer;
	typedef const T*	const_pointer;
	typedef T&			reference;
	typedef const T&	const_reference;
	typedef size_t		size_type;
	typedef ptrdiff_t	difference_type;

	// Ditto for this rebinding helper
	template <class U>
	struct rebind
	{
		typedef SecureAllocator<U, ScrubPolicy, AllocPolicy, 
			LockPolicy>	other;
	};

	// And these functions
	pointer			address(reference val) const		{ return &val; }
	const_pointer	address(const_reference val) const	{ return &val; }

	SecureAllocator()	throw()		{}
	template <class U>
	SecureAllocator(const SecureAllocator
		<U, ScrubPolicy, AllocPolicy, LockPolicy>&) throw()	{}
	~SecureAllocator() throw()	{}

	////////////////////////////////////////////////////////////////
	// The Standard requires allocators to implement max_count.
	// This duty should be handed off to AllocPolicy since it
	// handles the details of memory management.  However no one
	// calls this method in practice and neither should you.
	//
	// Reasons to ignore max_size:
	// - can't be trusted; another thread or process can reduce 
	//   available memory after check
	// - may vary with privileges; checking those is slow
	// - newer version of OS may support higher limits
	// - limits may not be enforced anyway (95/98)
	// - bounded Win32 heaps won't satisfy requests for more than
	//   0x7fff8 (2^20 - 8) bytes even if memory is available
	////////////////////////////////////////////////////////////////

	size_type max_size()
	{
		return getAllocator().max_count(sizeof(T));
	}

	// Redmond's STL uses non-standard signatures
#if defined(_MSC_VER)
	pointer allocate(size_type num, void* hint = 0)
#else
	pointer allocate(size_type num, const_pointer hint = 0)
#endif
	{
		typename LockPolicy::Lock lock;
		void* pmem = getAllocator().doAlloc(num * sizeof(T), hint);
		// don't rely on AllocPolicy to check for success
		if (! pmem)
		{
			throw std::bad_alloc();
		}
		return static_cast<pointer>(pmem);
	}

	// Redmond's STL needs this method to allocate memory for multiple
	// types in the absence of rebind.  Other platforms can ignore
	// this method.
#if defined(_MSC_VER)
	char* _Charalloc(size_type num)
	{
		typename LockPolicy::Lock lock;
		void* pmem = getAllocator().doAlloc(num * sizeof(T), NULL);
		// don't rely on AllocPolicy to check for success
		if (! pmem)
		{
			throw std::bad_alloc();
		}
		return static_cast<pointer>(pmem);
	}
#endif

#if defined(_MSC_VER)
	// Redmond's STL doesn't know how to rebind, so it uses  one
	// allocator to handle memory for multiple types.  Hence we
	// need to deallocate any type of memory.
	template <class U>
	void deallocate(U* pmem, size_type num)
	{
#else
	void deallocate(pointer pmem, size_type num)
	{
		// Make U a synonym for T so we can have one version of
		// the method body below.
		typedef	T	U;
#endif
		typename LockPolicy::Lock lock;
		getScrubber().onDeallocate(pmem, num * sizeof(U));
		getAllocator().doFree(pmem, num * sizeof(U));
	}

	void construct(pointer pmem, const value_type& val)
	{ 
		new(pmem) value_type(val);
	}

	void destroy(pointer pmem)
	{
		pmem->~value_type();
		getScrubber().onDestroy(pmem, sizeof(T));
	}

	// Return true if we can delete memory from the other allocator,
	// i.e. when they share the same policy classes
	template <class U>
	bool operator == (const SecureAllocator<U, 
		ScrubPolicy, AllocPolicy, LockPolicy>&)
	{
		return true;
	}

private:
	void operator=(const SecureAllocator&);

	AllocPolicy& getAllocator()
	{
		return PolicyHolder<AllocPolicy>::getPolicy();
	}

	ScrubPolicy& getScrubber()
	{
		return PolicyHolder<ScrubPolicy>::getPolicy();
	}
};

/*
template <class T, class U, class A, class S, class L>
bool operator == (const std::allocator<T>& a, 
				  const SecureAllocator<U, A, S, L>& s)
{
	return false;
}

template <class T, class U, class A, class S, class L>
bool operator == (const SecureAllocator<U, A, S, L>& s,
				  const std::allocator<T>& a)
{
	return false;
}
*/

////////////////////////////////////////////////////////////////
// STL-Port uses these template functions to rebind allocators 
// when there are no member templates (e.g. MSVC 6).
////////////////////////////////////////////////////////////////

#if defined (__SGI_STL) && !defined (_STLP_MEMBER_TEMPLATE_CLASSES)
#pragma message("Using STL-Port allocator rebinding functions")

_STLP_BEGIN_NAMESPACE

	template <class T, class U, class P1, class P2, class P3>
	inline SecureAllocator<U, P1, P2, P3>&
	__stl_alloc_rebind(SecureAllocator<T, P1, P2, P3>& a, const U*) 
	{  
		return (SecureAllocator<U, P1, P2, P3>&)(a); 
	}

	template <class T, class U, class P1, class P2, class P3>
	inline SecureAllocator<U, P1, P2, P3>
	__stl_alloc_create(const SecureAllocator<T, P1, P2, P3>& a, const U*)
	{
		return SecureAllocator<U, P1, P2, P3>();
	}

_STLP_END_NAMESPACE

#endif	// __SGI_STL

#endif	// SECALLOC_H
