/**************************************************************************
 * Copyright (c) 2002 Edward Elliott. All rights reserved.
 *
 * Various utility classes
 *************************************************************************/

#ifndef MISC_H
#define MISC_H

#include "ThreadPol.h"
#include <vector>
#include <algorithm>
#include <iterator>

/**
 * Holds a single policy instance.
 */
template <class Policy>
class PolicyHolder
{
public:
#ifndef	NO_BOOST_THREADS
	typedef MutexLocking<Policy*>::VolatileType		PtrType;
#else
	typedef Policy*	PtrType;
#endif

	static Policy& getPolicy()
	{
		static PtrType pPol = NULL;

		// Use double-checked locking pattern to ensure only 
		// one Policy object is created
		if (pPol == NULL)
		{
#ifndef	NO_BOOST_THREADS
			typename MutexLocking<PolicyHolder<Policy> >::Lock lock;
#endif
			if (pPol == NULL)
			{
				pPol = new Policy();
			}
		}
		return *pPol;
	}
};

/**
 * Generic implementation of max_count function.  Each AllocPolicy
 * could implement a more accurate version, taking into account
 * properties of the memory allocation scheme in use.  In practice
 * no one should call this method anyway, so most allocators can just
 * inherit this version.
 * See SecAlloc.h for a list of reasons max_count is ignored.
 */
class DummyMaxCount
{
public:
	size_t max_count(size_t objSize)	
	{
		// First version is correct according to the Standard, but
		// doesn't work with all compilers.  Second version is more
		// portable in practice, though not guaranteed to always work.
		
		//return std::numeric_limits<size_t>::max() / objSize;
		return size_t(-1) / objSize;
	}
};

/**
 * Keeps track of which pages are in use and which ones can be
 * unlocked.
 */
class PageTracker
{
public:
	// Add the given memory block to the in-use list
	void addBlock(void* pmem, size_t numBytes, size_t pageSize)
	{
		// Find the range of pages for this allocation
		size_t firstPage = size_t(pmem) / pageSize;
		size_t lastPage = (size_t(pmem) + numBytes) / pageSize;

		// Find the entry for firstPage, or where it belongs if none
		Tracker::iterator i = 
			std::lower_bound(tracker_.begin(), tracker_.end(), Page(firstPage));

		// For each page in the range, get/add the entry in PageTracker 
		// and increment the lockCnt
		for (size_t page = firstPage; page <= lastPage; ++page, ++i)
		{
			if (i == tracker_.end() || i->index_ != page)
			{
				i = tracker_.insert(i, Page(page));
			}
			++(i->lockCnt_);
		}
	}

	// Remove the memory block from the in-use list and unlock any
	// pages if possible.
	template<class UnlockPage>
	void removeBlock(void *pmem, size_t numBytes, size_t pageSize, UnlockPage unlockPage)
	{
		// Find the range of pages for this allocation
		size_t firstPage = size_t(pmem) / pageSize;
		size_t lastPage = (size_t(pmem) + numBytes) / pageSize;

		// Find the PageTracker entry for firstPage
		Tracker::iterator i = 
			std::lower_bound(tracker_.begin(), tracker_.end(), Page(firstPage));

		// For each page in the range, get/add the entry in PageTracker 
		// and decrement the lockCnt
		for (size_t page = firstPage; page <= lastPage; ++page)
		{
			if (i == tracker_.end() || i->index_ != page)
			{
				// No entry in tracker, we never allocated from this page.
				// Don't throw an exception, free will catch it later.
				continue;
			}
			--(i->lockCnt_);
			if (i->lockCnt_ == 0)
			{
				// No more allocations on this page, unlock it.
				// Range is 1 (byte) to only unlock a single page.
				unlockPage(pmem, 1);
			}
		}
	}

private:
	// Usage info on a page
	struct Page
	{
		explicit Page(size_t idx)
		{
			index_ = idx;
			lockCnt_ = 0;
		}

		size_t index_;
		size_t lockCnt_;

		// For sorting
		bool operator < (const Page& rhs) const	{ return index_ < rhs.index_; }
	};

	typedef std::vector<Page>	Tracker;
	Tracker	tracker_;
};

#endif	// MISC_H
